""" axishell module
to create x-axisymmetric shells for scientific computations
"""

import warnings

import numpy as np

from kokiy import Shell


class CartShell(Shell):

    def __init__(self, *args, **kwargs):
        '''
        A workaround to keep all the signatures available in arnica. To be
        removed when backward compatibility is no longer an issue.
        '''
        if type(args[2]) is np.ndarray:
            self._init_with_ctrl_pts(*args, **kwargs)
        else:
            self._init(*args)

    def _init(self, n_trans, n_longi, zero, umax, vmax):
        # pylint: disable=too-many-instance-attributes
        r"""\
        *Base class for cartesian computationnal shells*

        CartShell class is based on Shell class, inheriting of Shell methods.
        At initialization, the shell is builded an an cartesian shell.

        Some attributes are based on **u** and **v**, defined as the curvilinear
        longitudinal (u) abscissa and curvilinear transversal (v) abscissa
        respectively.

        The shell is set using three corners (assuming they are othogonal)

        ::
              X umax
              ________________
              |              |            Cartesian system
              |              |            Example longi/u <=> x  and v <==> z
              |              |            zero. (0, 0, 0)
              |              |            umax. (1, 0, 0)
              |              |            vmax. (0, 0, 1)
              |              |            x - longi/u
              |              |            ^
              |              |            |
              |              |            |
              |              |            O-----> z - transvers/v
              |______________| X vmax     y
              X zero.

        :param n_transvers: Number of azimuthal shell points
        :type n_transvers: int
        :param n_longi: Number of longitudinal shell points
        :type n_longi: int
        :param z_range: Extrusion range in transvers direction
        :type z_range: float
        :param zero: Array of dim (3) lower left corner of shell
        :param umax: Array of dim (3) corner of shell for umax
        :param vmax: Array of dim (3) corner of shell for vmax

        Public attributes :

            - **shape** - Shape of the shell : (n_transvers, n_longi)
            - **width_matrix** - Dict containing thickness matrices of shape 'shape'
            - **matrix** - Dict containing fields of shape 'shape' - To deprecate

        Private attributes :

            - **_xyz**
            - **_rad**
            - **_theta**
            - **_n_x**
            - **_n_r**
            - **_n_y**
            - **_n_z**
            - **_du**
            - **_dv**
            - **_abs_curv**
            - **_dwu**
            - **_dwv**
            - **_surf**
        """
        # TODO: rewrite docstrings (too long and prone to rapidly be outdated)
        super().__init__(n_trans, n_longi)
        self._build_shell(
            np.array(zero),
            np.array(umax),
            np.array(vmax))

    def _build_shell(self, zero, umax, vmax):
        # pylint: disable=arguments-differ
        """
        *Build shell from geometric features*

            - Construct a spline used as base for extrusion\
              from control points : tck
            - Discretise the spline : shell_crest
            - Compute normal vectors for the 1D shell_crest
            - Compute r,n_x,n_r-components for 2D shell
            - Compute theta-components for 2D shell
            - Compute xyz,n_y,n_z-components for 2D shell
        """

        vec_u = umax - zero
        vec_v = vmax - zero

        norm = np.cross(vec_u, vec_v)
        norm /= np.linalg.norm(norm)

        u_range = np.linspace(0., 1., num=self.shape[0])
        v_range = np.linspace(0., 1., num=self.shape[1])
        d_u = np.linalg.norm(vec_u) / (self.shape[0] - 1)
        d_v = np.linalg.norm(vec_v) / (self.shape[1] - 1)

        v_coor, u_coor = np.meshgrid(v_range, u_range)
        # multiply coord matrix by each vec component
        u_coor = vec_u[:, np.newaxis, np.newaxis] * u_coor[np.newaxis, :, :]
        v_coor = vec_v[:, np.newaxis, np.newaxis] * v_coor[np.newaxis, :, :]
        # TODO: use axis=0 instead? seems more intuitive (avoid np.take)
        self._xyz = np.stack([zero[i] + u_coor[i] + v_coor[i] for i in range(3)],
                             axis=-1)

        # Compute radius and azimuth matrix of shape (n_transvers, n_longi)
        # TODO: avoid using take
        self._rad = np.sqrt(
            np.take(self._xyz, 1, -1) ** 2
            + np.take(self._xyz, 2, -1) ** 2)
        self._theta = np.arctan2(
            np.take(self._xyz, 2, -1),
            np.take(self._xyz, 1, -1))

        # Compute x,y,z,r-normal matrix of shape (n_transvers, n_longi)
        self._n_x = np.full(self.shape, norm[0])
        self._n_y = np.full(self.shape, norm[1])
        self._n_z = np.full(self.shape, norm[2])
        self._n_r = self._n_y * np.cos(self._theta)

        self._du = np.full(self.shape, d_u)
        self._dv = np.full(self.shape, d_v)

        # Compute abs_curv array of shape (n_longi,)
        self._abs_curv = np.cumsum(np.take(self._du, 0, 0)) - self._du[0, 0]

        # Compute weight intervals in u and v directions array of shape (n_transvers, n_longi)
        self._dwu = self._du.copy()
        self._dwu[:, (0, -1)] /= 2
        self._dwv = self._dv.copy()
        self._dwv[(0, -1), :] /= 2

        # Compute surface nodoes array of shape (n_transvers, n_longi)
        self._surf = np.multiply(self._dwu, self._dwv)

    def _init_with_ctrl_pts(self, n_trans, n_longi, z_range, ctrl_pts_x,
                            ctrl_pts_y, z_min=None):
        # pylint: disable=too-many-instance-attributes
        r"""\
        *Base class for cartesian computationnal shells*

        CartShell class is based on Shell class, inheriting of Shell methods.
        At initialization, the shell is builded an an cartesian shell parallel to z.

        Some attributes are based on **u** and **v**, defined as the curvilinear\
        longitudinal (x/y) abscissa and curvilinear transversal (z) abscissa\
        respectively.

        ::
                       (x_n, y_n)
              ________X________
              |       |       |            Cartesian system
              |       |       |            Example longi/u <=> x
              |       |       |
              |       |       |            x - longi/u
              |    <----->    |            ^
              |       |       |            |
              |       |       |            |
              |       |       |            O-----> z - transvers/v
              |_______X_______|           y
                       (x_0, y_0)

        :param n_transvers: Number of azimuthal shell points
        :type n_transvers: int
        :param n_longi: Number of longitudinal shell points
        :type n_longi: int
        :param z_range: Extrusion range in transvers direction
        :type z_range: float
        :param ctrl_pts_x: Array of dim (n,) of x-coordinates of the points defining the spline
        :param ctrl_pts_y: Array of dim (n,) of y-coordinates of the points defining the spline

        Optionals arguments :
        :param z_min: Minimum transversal value of the shell
        :type z_min: float

        Public attributes :

            - **shape** - Shape of the shell : (n_transvers, n_longi)
            - **width_matrix** - Dict containing thickness matrices of shape 'shape'
            - **matrix** - Dict containing fields of shape 'shape' - To deprecate

        Private attributes :

            - **_xyz**
            - **_rad**
            - **_theta**
            - **_n_x**
            - **_n_r**
            - **_n_y**
            - **_n_z**
            - **_du**
            - **_dv**
            - **_abs_curv**
            - **_dwu**
            - **_dwv**
            - **_surf**
        """
        warnings.warn('This signature will be deprecated. Please, use alternative.',
                      DeprecationWarning)

        super().__init__(n_trans, n_longi)

        self.ctrl_pts_x = ctrl_pts_x
        self.ctrl_pts_y = ctrl_pts_y
        self.z_range = z_range
        self.z_min = z_min

        self._build_shell_with_ctrl_pts()

    def _build_shell_with_ctrl_pts(self):
        # pylint: disable=arguments-differ
        """
        *Build shell from geometric features*

            - Construct a spline used as base for extrusion\
              from control points : tck
            - Discretise the spline : shell_crest
            - Compute normal vectors for the 1D shell_crest
            - Compute r,n_x,n_r-components for 2D shell
            - Compute theta-components for 2D shell
            - Compute xyz,n_y,n_z-components for 2D shell
        """

        # Construct Shell Crest
        shell_crest = self._compute_shell_crest(self.ctrl_pts_x, self.ctrl_pts_y)

        # Compute xyz matrix of shape (n_transvers, n_longi, 3)
        tmp_x = np.tile(shell_crest[0], (self.shape[0], 1))
        tmp_y = np.tile(shell_crest[1], (self.shape[0], 1))

        trans_z = 0.
        if self.z_min is not None:
            trans_z = 0.5 * self.z_range + self.z_min
        min_z = trans_z - 0.5 * self.z_range
        max_z = trans_z + 0.5 * self.z_range
        tmp_z = np.transpose(np.tile(np.linspace(min_z,
                                                 max_z,
                                                 num=self.shape[0]),
                                     (self.shape[1], 1)))
        self._xyz = np.stack((tmp_x, tmp_y, tmp_z), axis=-1)

        # Compute radius and azimuth matrix of shape (n_transvers, n_longi)
        self._rad = np.sqrt(np.take(self._xyz, 1, -1) ** 2
                            + np.take(self._xyz, 2, -1) ** 2)
        self._theta = np.arctan2(np.take(self._xyz, 2, -1),
                                 np.take(self._xyz, 1, -1))

        # Compute x,y,z,r-normal matrix of shape (n_transvers, n_longi)
        xy_nml_1d = self._compute_shellcrest_nml(shell_crest)
        self._n_x = np.tile(xy_nml_1d[0], (self.shape[0], 1))
        self._n_y = np.tile(xy_nml_1d[1], (self.shape[0], 1))
        self._n_z = np.zeros(self.shape)
        self._n_r = self._n_y * np.cos(self._theta)

        # Compute du,dv matrix of shape (n_transvers, n_longi)
        self._du = np.pad(np.sqrt(np.diff(self._xyz[:, :, 0], axis=1) ** 2
                                  + np.diff(self._xyz[:, :, 1], axis=1) ** 2),
                          ((0, 0), (1, 0)),
                          'edge')
        self._dv = np.pad(np.diff(self._xyz[:, :, 2], axis=0),
                          ((1, 0), (0, 0)),
                          'edge')
        # Compute abs_curv array of shape (n_longi,)
        self._abs_curv = np.cumsum(np.take(self._du, 0, 0)) - self._du[0, 0]

        # Compute weight intervals in u and v directions array of shape (n_transvers, n_longi)
        self._dwu = self._du.copy()
        self._dwu[:, (0, -1)] /= 2
        self._dwv = self._dv.copy()
        self._dwv[(0, -1), :] /= 2

        # Compute surface nodoes array of shape (n_transvers, n_longi)
        self._surf = np.multiply(self._dwu, self._dwv)



"""
Module for cartesian or axi-cylindrical cut plane tools
"""

import numpy as np

from arnica.utils.vector_actions import yz_to_theta
from arnica.utils.cloud2cloud import cloud2cloud

from kokiy import AxiShell
from kokiy import CartShell


# TODO: do we really care with those that misuse import?
__all__ = ["infer_shell_from_xyz", "axishell_create_structured_base",
           "cartshell_create_structured_base",
           "interpolate_solutions_on_shell",
           "shell_geom_type", "shell_repr"]


def infer_shell_from_xyz(mesh_xyz, geom_type, shape, bnd_angles=None):
    """ Build a shell from an AVBP Cut

    :param geom_type: str, either "cart" or "axicyl".
    :param shape: tuple (nu,nv), discretization of the shell
    :param bnd_angles: Boundary angles to keep a fraction of the cuts
    :type bnd_angles: list of 2 floats to limite the range of azimuth
    """

    #mesh = find_mesh_for_sol(paths, prefix)
    # Build shell
    if geom_type == "axicyl":
        shell = axishell_create_structured_base(
            shape[0],
            shape[1],
            mesh_xyz,
            bnd_angles)
    elif geom_type == "cart":
        shell = cartshell_create_structured_base(
            shape[0],
            shape[1],
            mesh_xyz)
    return shell


def axishell_create_structured_base(n_longi, n_azi, mesh, bnd_angles=None):
    """
    *Obtain the base in structured mesh*

    Assumes the mesh is the x_axis rotated extrusion of an (x,r) curve.

    :param n_longi: Number of longitudinal points of the shell
    :type n_longi: int
    :param n_azi: Number of azimuthal points of the shell
    :type n_azi: int
    :param mesh: Array of coordinates of the cut mesh of shape (n, 3)
    :param bnd_angles: Boundary angles to keep a fraction of the cuts
    :type bnd_angles: list of 2 floats

    :returns:
        - **shell** -  An array representing axysymmetric computational shell
    """

    x_vals = mesh[:, 0]
    r_vals = np.sqrt(np.square(mesh[:, 1]) + np.square(mesh[:, 2]))

    corners_idx = [np.argmin(x_vals), np.argmax(x_vals)]
    if corners_idx[0] == corners_idx[1]:
        corners_idx = [np.argmin(r_vals), np.argmax(r_vals)]
    angle_vals = np.rad2deg(yz_to_theta(mesh))
    angle_range = np.max(angle_vals) - np.min(angle_vals)
    angle_min = np.amin(angle_vals)

    if bnd_angles is not None:
        try:
            if bnd_angles[1] - bnd_angles[0] > angle_range:
                raise ValueError
        except ValueError as value_error:
            msg = "Your cuts seem smaller than the range of angles you specified:\n"
            msg += f"{bnd_angles[1]} - {bnd_angles[0]} > {angle_range}\n"
            msg += "Please check the input file."
            raise ValueError(msg) from value_error
        try:
            if bnd_angles[0] < np.amin(angle_vals) or bnd_angles[1] > np.amax(angle_vals):
                raise ValueError
        except ValueError as value_error:
            msg = "The bounds you specified are outside the actual bounds of the cuts.\n"
            msg += "Please check the input file"
            raise ValueError(msg) from value_error
        angle_range = bnd_angles[1] - bnd_angles[0]
        angle_min = bnd_angles[0]

    shell = AxiShell(
        n_azi,
        n_longi,
        angle_range,
        x_vals[corners_idx],
        r_vals[corners_idx],
        angle_min,
    )

    return shell


def cartshell_create_structured_base(n_longi, n_trans, mesh):
    # pylint: disable=invalid-name
    """
    *Obtain the base in structured mesh*

    Assumes the path is square, aligned with y and z, with x = constant.

    :param n_longi: Number of longitudinal points of the shell
    :type n_longi: int
    :param n_trans: Number of transversal points of the shell
    :type n_trans: int
    :param mesh: Array of coordinates of the cut mesh of shape (n, 3)

    :returns:
        - **shell** -  An array representing cartesian computational shell
    """

    x_vals, y_vals, z_vals = mesh.T

    zero = (x_vals.mean(), y_vals.min(), z_vals.min())
    v_max = (x_vals.mean(), y_vals.max(), z_vals.min())
    u_max = (x_vals.mean(), y_vals.min(), z_vals.max())
    shell = CartShell(
        n_trans,
        n_longi,
        zero,
        u_max,
        v_max,
    )

    return shell


# TODO :  Eloi, these two functions are the thing to optimize for you. Here, and in cloud2cloud...

def _shell_interp_on_struct_base(shell, mesh, raw_data):
    """
    *Interpolate unstructured data on a structured grid*

    :param shell: AxiShell or CartShell object
    :param mesh: Array of shape (N,3) containing coordinates of raw_data
    :param raw_data: A numpy structured array gathering the unstructured data

    :returns: **struct_data** -  An array holding the intrepolated structured data
    """

    dim = shell.xyz.shape[-1]
    target_xyz = shell.xyz.reshape(-1, dim)
    raw_data_dict = {key: raw_data[key] for key in raw_data.dtype.names}  # Structured array to dict

    struct_dict = cloud2cloud(mesh, raw_data_dict, target_xyz, stencil=4, power=4)

    struct_data = np.empty(shell.shape, dtype=raw_data.dtype)
    for key in struct_dict:
        struct_data[key] = struct_dict[key].reshape(shell.shape)

    return struct_data


def interpolate_solutions_on_shell(shell, mesh, raw_data):
    """
    *Loads AVBP solutions interpolated on structured mesh.*

    :param shell: A shell
    :param data_dict: | A dict( ) containing data on the format \
                      **data_dict** [*key*] = array(*n_steps*, *n_theta*, *n_r*)
                      with various names "rho", "rhou"...

    :returns:
        - **interpol_data** - A dict( ) containing interpolated data
    """

    # TODO : make this interpolation without the for on times
    # probably inside the cloud to cloud
    n_steps = raw_data.shape[0]
    interpol_data = np.zeros((n_steps, *shell.shape), dtype=raw_data.dtype)
    for step in range(n_steps):
        interpol_data[step] = _shell_interp_on_struct_base(
            shell,
            mesh,
            raw_data[step],
        )

    return interpol_data


def shell_geom_type(shell):
    """Return the shell geom_type"""
    # TODO: really required?
    if isinstance(shell, AxiShell):
        return "axicyl"
    elif isinstance(shell, CartShell):
        return "cart"
    else:
        raise RuntimeError("Shell object must be an AxiShell or a CartShell")


def shell_repr(shell):
    """return a shell description"""
    # TODO: why not to make it a method of shell?
    out = f"""
Shell
-----

shape :{shell.shape}
type : {shell_geom_type(shell)}

"""
    return out
